#!/usr/bin/env python

"""
    conductiviy.py: Module is used to implement Earth conductivity methods
"""

__author__ = "Murphy, B.; Chakraborty, S."
__copyright__ = ""
__credits__ = []
__license__ = "MIT"
__version__ = "1.0."
__maintainer__ = "Chakraborty, S."
__email__ = "shibaji7@vt.edu"
__status__ = "Research"

import os

import numpy as np
import pandas as pd
from loguru import logger
from pyproj import Geod
from scipy.interpolate import RegularGridInterpolator
from scipy.io import netcdf_file

from scubas.datasets import Site
from scubas.utils import RecursiveNamespace


class ConductivityProfile(object):
    """
    Class is dedicated to create conductivity profiles
    from LITHO1.0 model.

    1.  LITHO1.0 as the basis for constructing the resistivity profiles
        http://ds.iris.edu/ds/products/emc-litho10/
    """

    def __init__(
        self,
        conductivity_params=dict(
            earth_model="LITHO1.0.nc",
            grid_interpolation_method="nearest",
            seawater_resistivity=0.3,
            sediment_resistivity=3.0,
            crust_resistivity=3000.0,
            lithosphere_resistivity=1000.0,
            asthenosphere_resistivity=100.0,
            transition_zone_resistivity=10.0,
            lower_mantle_resistivity=1.0,
            transition_zone_top=410.0,
            transition_zone_bot=660.0,
            profile_max_depth=1000.0,
            uri="http://ds.iris.edu/files/products/emc/emc-files/LITHO1.0.nc",
        ),
    ):
        self.cprop = RecursiveNamespace(**conductivity_params)
        self.earth_model = self.cprop.earth_model

        # NOTE these are specified in terms of resistivity, in ohm-m
        self.seawater_resistivity = self.cprop.seawater_resistivity
        self.sediment_resistivity = self.cprop.sediment_resistivity
        self.crust_resistivity = self.cprop.crust_resistivity
        self.lithosphere_resistivity = self.cprop.lithosphere_resistivity
        self.asthenosphere_resistivity = self.cprop.asthenosphere_resistivity
        self.transition_zone_resistivity = self.cprop.transition_zone_resistivity
        self.lower_mantle_resistivity = self.cprop.lower_mantle_resistivity

        # fixed layer depths (below the lithosphere), in km
        self.transition_zone_top = self.cprop.transition_zone_top
        self.transition_zone_bot = self.cprop.transition_zone_bot
        self.profile_max_depth = self.cprop.profile_max_depth

        # things that control how this code works in detail...
        # Options: "nearest" or "linear"
        self.grid_interpolation_method = self.cprop.grid_interpolation_method

        # Load netcdf file
        self.load_earth_model()
        return

    def load_earth_model(self):
        """
        Returns a dict containing the pertinent parts of the Earth model
        Dict entries are callable interpolation objects
        """
        os.makedirs(".scubas_config/", exist_ok=True)
        filename = ".scubas_config/" + self.earth_model
        if not os.path.exists(filename):
            uri = self.cprop.uri
            os.system(f"wget -O {filename} {uri}")
        with netcdf_file(filename) as f:
            latitude = np.copy(f.variables["latitude"][:])
            longitude = np.copy(f.variables["longitude"][:])
            # base of lithosphere (top of asthenosphere)
            asthenospheric_mantle_top_depth = np.copy(
                f.variables["asthenospheric_mantle_top_depth"][:]
            )
            # top/bottom of mantle lithosphere
            lithosphere_bottom_depth = np.copy(f.variables["lid_bottom_depth"][:])
            lithosphere_top_depth = np.copy(f.variables["lid_top_depth"][:])
            # crustal layers
            lower_crust_bottom_depth = np.copy(
                f.variables["lower_crust_bottom_depth"][:]
            )
            lower_crust_top_depth = np.copy(f.variables["lower_crust_top_depth"][:])
            middle_crust_bottom_depth = np.copy(
                f.variables["middle_crust_bottom_depth"][:]
            )
            middle_crust_top_depth = np.copy(f.variables["middle_crust_top_depth"][:])
            upper_crust_bottom_depth = np.copy(
                f.variables["upper_crust_bottom_depth"][:]
            )
            upper_crust_top_depth = np.copy(f.variables["upper_crust_top_depth"][:])
            # sediment layers
            lower_sediments_bottom_depth = np.copy(
                f.variables["lower_sediments_bottom_depth"][:]
            )
            lower_sediments_top_depth = np.copy(
                f.variables["lower_sediments_top_depth"][:]
            )
            middle_sediments_bottom_depth = np.copy(
                f.variables["middle_sediments_bottom_depth"][:]
            )
            middle_sediments_top_depth = np.copy(
                f.variables["middle_sediments_top_depth"][:]
            )
            upper_sediments_bottom_depth = np.copy(
                f.variables["upper_sediments_bottom_depth"][:]
            )
            upper_sediments_top_depth = np.copy(
                f.variables["upper_sediments_top_depth"][:]
            )
            # water levels
            water_bottom_depth = np.copy(f.variables["water_bottom_depth"][:])
            water_top_depth = np.copy(f.variables["water_top_depth"][:])

        self.lithosphere_model = {
            "latitude": latitude,
            "longitude": longitude,
            "asthenospheric_mantle_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                asthenospheric_mantle_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "lithosphere_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                lithosphere_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "lithosphere_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                lithosphere_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "lower_crust_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                lower_crust_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "lower_crust_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                lower_crust_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "middle_crust_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                middle_crust_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "middle_crust_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                middle_crust_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "upper_crust_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                upper_crust_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "upper_crust_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                upper_crust_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "lower_sediments_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                lower_sediments_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "lower_sediments_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                lower_sediments_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "middle_sediments_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                middle_sediments_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "middle_sediments_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                middle_sediments_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "upper_sediments_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                upper_sediments_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "upper_sediments_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                upper_sediments_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "water_bottom_depth": RegularGridInterpolator(
                (latitude, longitude),
                water_bottom_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
            "water_top_depth": RegularGridInterpolator(
                (latitude, longitude),
                water_top_depth,
                method=self.grid_interpolation_method,
                bounds_error=True,
            ),
        }
        return

    def get_interpolation_points(self, pt0, pt1):
        """ """
        globe = Geod(ellps="WGS84")

        # somewhat janky way of figuring out how many interp points to have between
        # the bin edges... basic idea is that we want one point spaced every ~1 deg,
        # since that"s the resolution of LITHO1.0
        npts = int(round(np.sqrt((pt1[1] - pt0[1]) ** 2.0 + (pt1[0] - pt0[0]) ** 2.0)))
        if npts < 1:
            npts = 1

        # obtain equally spaced points along a geodesic (doesn"t include end points)
        latlons = globe.npts(pt0[1], pt0[0], pt1[1], pt1[0], npts)

        # add the bin edges to the lat/lon list... note this array is now [lon, lat]
        interpolation_points = np.vstack(
            (np.array([pt0[1], pt0[0]]), np.array(latlons), np.array([pt1[1], pt1[0]]))
        )

        # now swap back to [lat, lon]
        interpolation_points = np.vstack(
            (interpolation_points[:, 1], interpolation_points[:, 0])
        ).T

        return interpolation_points

    def get_water_layer(self, lithosphere_model, pts):
        """
        Interpolates and averages seawater layer thickness within the given bin
        """

        water_top_values = lithosphere_model["water_top_depth"](pts)
        water_bot_values = lithosphere_model["water_bottom_depth"](pts)

        water_top = np.nanmean(water_top_values)
        water_bot = np.nanmean(water_bot_values)
        water_thk = water_bot - water_top

        # sanity check... water_top can be slightly different than zero,
        # but shouldn"t be that much different
        if np.absolute(water_top) > 0.01:
            logger.warning(f"PROBLEM: water_top doesn't make sense: {water_top}")

        return water_thk

    def get_sediment_layer(self, lithosphere_model, pts):
        """
        Interpolates and averages sediment layer thickness within the given bin
        """
        upper_sed_top_values = lithosphere_model["upper_sediments_top_depth"](pts)
        upper_sed_bot_values = lithosphere_model["upper_sediments_bottom_depth"](pts)
        upper_sed_thk_values = upper_sed_bot_values - upper_sed_top_values

        middle_sed_top_values = lithosphere_model["middle_sediments_top_depth"](pts)
        middle_sed_bot_values = lithosphere_model["middle_sediments_bottom_depth"](pts)
        middle_sed_thk_values = middle_sed_bot_values - middle_sed_top_values

        lower_sed_top_values = lithosphere_model["lower_sediments_top_depth"](pts)
        lower_sed_bot_values = lithosphere_model["lower_sediments_bottom_depth"](pts)
        lower_sed_thk_values = lower_sed_bot_values - lower_sed_top_values

        sed_thk_values = np.nansum(
            np.vstack(
                (upper_sed_thk_values, middle_sed_thk_values, lower_sed_thk_values)
            ),
            axis=0,
        )

        sed_thk = np.nanmean(sed_thk_values)

        return sed_thk

    def get_crust_layer(self, lithosphere_model, pts):
        """
        Interpolates and averages crust layer thickness within the given bin
        """
        upper_crust_top_values = lithosphere_model["upper_crust_top_depth"](pts)
        upper_crust_bot_values = lithosphere_model["upper_crust_bottom_depth"](pts)
        upper_crust_thk_values = upper_crust_bot_values - upper_crust_top_values

        middle_crust_top_values = lithosphere_model["middle_crust_top_depth"](pts)
        middle_crust_bot_values = lithosphere_model["middle_crust_bottom_depth"](pts)
        middle_crust_thk_values = middle_crust_bot_values - middle_crust_top_values

        lower_crust_top_values = lithosphere_model["lower_crust_top_depth"](pts)
        lower_crust_bot_values = lithosphere_model["lower_crust_bottom_depth"](pts)
        lower_crust_thk_values = lower_crust_bot_values - lower_crust_top_values

        crust_thk_values = np.nansum(
            np.vstack(
                (
                    upper_crust_thk_values,
                    middle_crust_thk_values,
                    lower_crust_thk_values,
                )
            ),
            axis=0,
        )

        crust_thk = np.nanmean(crust_thk_values)

        return crust_thk

    def get_lithosphere_layer(self, lithosphere_model, pts):
        """
        Interpolates and averages mantle lithosphere thickness within the given bin
        """

        lithosphere_top_values = lithosphere_model["lithosphere_top_depth"](pts)
        lithosphere_bot_values = lithosphere_model["lithosphere_bottom_depth"](pts)
        asthenosphere_top_values = lithosphere_model["asthenospheric_mantle_top_depth"](
            pts
        )

        litho_top = np.nanmean(lithosphere_top_values)
        litho_bot = np.nanmean(lithosphere_bot_values)
        litho_thk = litho_bot - litho_top
        astheno_top = np.nanmean(asthenosphere_top_values)

        # sanity check... make sure top of asthenosphere is the same as bottom of the lithosphere
        if litho_bot != astheno_top:
            logger.warning(
                f"PROBLEM: lithosphere-asthenosphere dont line up: {litho_bot}, {astheno_top}"
            )

        return litho_thk

    def get_upper_mantle_layer(self, lithosphere_model, pts):
        """
        Interpolates and averages upper mantle thickness within the given bin
        """
        asthenosphere_top_values = lithosphere_model["asthenospheric_mantle_top_depth"](
            pts
        )
        astheno_top = np.nanmean(asthenosphere_top_values)
        astheno_thk = self.transition_zone_top - astheno_top
        return astheno_thk

    def get_transition_zone_layer(self):
        """
        Return fixed mantle transition zone thickness
        """
        return self.transition_zone_bot - self.transition_zone_top

    def get_lower_mantle_layer(self):
        """
        Return fixed lower mantle thickness
        """
        return self.profile_max_depth - self.transition_zone_bot

    def _compile_profile_(self, pts):
        """
        Compile profile for a list of points
        """
        # now progress through the model layers from near-surface to deep Earth...
        # first the water layer
        water_thk = self.get_water_layer(self.lithosphere_model, pts)
        # sediment layer
        sed_thk = self.get_sediment_layer(self.lithosphere_model, pts)
        # crust layer
        crust_thk = self.get_crust_layer(self.lithosphere_model, pts)
        # mantle lithosphere layer
        litho_thk = self.get_lithosphere_layer(self.lithosphere_model, pts)
        # asthenosphere layer
        astheno_thk = self.get_upper_mantle_layer(self.lithosphere_model, pts)
        # transition zone layer
        tz_thk = self.get_transition_zone_layer()
        # lower mantle layer
        lm_thk = self.get_lower_mantle_layer()
        resistivity_profile = np.array(
            [
                [water_thk, self.seawater_resistivity, "Seawater"],
                [sed_thk, self.sediment_resistivity, "Sediment"],
                [crust_thk, self.crust_resistivity, "Crust"],
                [litho_thk, self.lithosphere_resistivity, "Lithosphere"],
                [astheno_thk, self.asthenosphere_resistivity, "Upper Mantle"],
                [tz_thk, self.transition_zone_resistivity, "Transition Zone"],
                [lm_thk, self.lower_mantle_resistivity, "Lower Mantle"],
            ]
        )
        rf = pd.DataFrame()
        rf["thickness"], rf["resistivity"], rf["name"] = (
            np.array(resistivity_profile[:, 0]).astype(float),
            resistivity_profile[:, 1],
            resistivity_profile[:, 2],
        )
        return rf

    @staticmethod
    def compile_profile(
        latlon,
        kind="rounded",
        to_site=True,
        site_name="",
        site_description="",
        **conductivity_params,
    ):
        """
        Compile profiles for a set of latlons
        """
        cp = ConductivityProfile(**conductivity_params)
        if kind == "rounded":
            latlon = np.rint(latlon)
        logger.info(f"Lat-lon: {latlon}")
        profile = cp._compile_profile_(latlon)
        logger.info(f"Compiled Profile \n {profile}")
        if to_site:
            profile = Site.init(
                1.0 / profile["resistivity"].to_numpy(dtype=float),
                profile["thickness"].to_numpy(dtype=float),
                profile["name"],
                site_description,
                site_name,
            )
        return profile

    @staticmethod
    def compile_profiles(
        latlons,
        kind="rounded",
        to_site=True,
        site_names=[],
        site_descriptions=[],
        **conductivity_params,
    ):
        """
        Compile profiles for a set of latlons
        """

        cp = ConductivityProfile(**conductivity_params)
        profiles = []
        for i, latlon in enumerate(latlons):
            if kind == "rounded":
                latlon = np.rint(latlon)
            logger.info(f"Lat-lon: {latlon}")
            profile = cp._compile_profile_(latlon)
            logger.info(f"Compiled Profile \n {profile}")
            if to_site:
                site_name = site_names[i] if i < len(site_names) else ""
                site_description = (
                    site_descriptions[i] if i < len(site_descriptions) else ""
                )
                profile = Site.init(
                    1.0 / profile["resistivity"].to_numpy(dtype=float),
                    profile["thickness"].to_numpy(dtype=float),
                    profile["name"],
                    site_description,
                    site_name,
                )
            profiles.append(profile)
        return profiles

    @staticmethod
    def compile_bined_profile(
        bined_latlon,
        to_site=True,
        site_name="",
        site_description="",
        **conductivity_params,
    ):
        """
        Compile profiles for a binned latlons
        """
        cp = ConductivityProfile(**conductivity_params)
        ipts = cp.get_interpolation_points(bined_latlon[0], bined_latlon[1])
        profile = cp._compile_profile_(ipts)
        logger.info(f"Compiled Profile \n {profile}")
        if to_site:
            profile = Site.init(
                1.0 / profile["resistivity"].to_numpy(dtype=float),
                profile["thickness"].to_numpy(dtype=float),
                profile["name"],
                site_description,
                site_name,
            )
        return profile

    @staticmethod
    def compile_bined_profiles(
        bined_latlons,
        to_site=True,
        site_names=[],
        site_descriptions=[],
        **conductivity_params,
    ):
        """
        Compile profiles for a binned latlons
        """
        cp = ConductivityProfile(**conductivity_params)
        profiles = []
        nbins = len(bined_latlons) - 1
        for i in range(nbins):
            ipts = cp.get_interpolation_points(
                bined_latlons[i, :], bined_latlons[i + 1, :]
            )
            profile = cp._compile_profile_(ipts)
            profile.thickness = profile.thickness * 1e3
            logger.info(f"Compiled Profile \n {profile}")
            if to_site:
                site_name = site_names[i] if i < len(site_names) else ""
                site_description = (
                    site_descriptions[i] if i < len(site_descriptions) else ""
                )
                profile = Site.init(
                    1.0 / profile["resistivity"].to_numpy(dtype=float),
                    profile["thickness"].to_numpy(dtype=float),
                    profile["name"],
                    site_description,
                    site_name,
                )
            profiles.append(profile)
        return profiles

    @staticmethod
    def compile_mcmc_bined_profiles(
        bined_latlons,
        n=100,
        continental_shelves_thickness_range=[0.01, 1.0],
        to_site=True,
        site_names=[],
        site_descriptions=[],
        random_seed=0,
        **conductivity_params,
    ):
        """
        Compile MCMC profiles for a binned latlons
        """
        np.random.seed(random_seed)
        cp = ConductivityProfile(**conductivity_params)
        mcmc_profiles, profiles = [], []
        for i, latlon in enumerate(bined_latlons):
            profile = cp._compile_profile_(latlon)
            profile.fillna(0, inplace=True)
            profiles.append(profile)
        for _ in range(n):
            mc_profiles = []
            for i in range(len(profiles) - 1):
                profile = profiles[i].copy()
                profile.thickness = (
                    np.random.uniform(
                        np.array(profiles[i].thickness),
                        np.array(profiles[i + 1].thickness),
                    )
                    * 1e3
                )
                if (i == 0) or (i == len(profiles) - 2):
                    profile.thickness[0] = np.random.uniform(
                        continental_shelves_thickness_range[0],
                        continental_shelves_thickness_range[1],
                    )
                logger.info(f"Compiled Profile \n {profile}")
                if to_site:
                    site_name = site_names[i] if i < len(site_names) else ""
                    site_description = (
                        site_descriptions[i] if i < len(site_descriptions) else ""
                    )
                    profile = Site.init(
                        1.0 / profile["resistivity"].to_numpy(dtype=float),
                        profile["thickness"].to_numpy(dtype=float),
                        profile["name"],
                        site_description,
                        site_name,
                    )
                mc_profiles.append(profile)
            mcmc_profiles.append(mc_profiles)
        return mcmc_profiles



#!/usr/bin/env python

import argparse
import os
import shutil


def clean():
    if os.path.exists("dist/"):
        shutil.rmtree("dist/")
    if os.path.exists("build/"):
        shutil.rmtree("build/")
    if os.path.exists("scubas.egg-info/"):
        shutil.rmtree("scubas.egg-info/")
    os.system("find . -type d -name '.ipynb_checkpoints' -exec rm -rf {} +")
    os.system("find . -type d -name '__pycache__' -exec rm -rf {} +")
    return


def build():
    os.system("isort -rc -sl .")
    os.system("autoflake --remove-all-unused-imports -i -r .")
    os.system("isort -rc -m 3 .")
    os.system("black .")
    os.system("python setup.py sdist bdist_wheel")
    return


def uplaod_pip():
    clean()
    build()
    os.system("twine upload dist/* --verbose")
    return


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b", "--build", action="store_true", help="Build the project from scratch"
    )
    parser.add_argument(
        "-rm", "--clean", action="store_true", help="Cleaning pre-existing files"
    )
    parser.add_argument(
        "-rb",
        "--rebuild",
        action="store_true",
        help="Clean and rebuild the project from scratch",
    )
    parser.add_argument(
        "-upip",
        "--upip",
        action="store_true",
        help="Upload code to PIP repository",
    )
    args = parser.parse_args()
    if args.clean:
        clean()
    if args.build:
        build()
    if args.rebuild:
        clean()
        build()
    if args.upip:
        uplaod_pip()



"""
    currents.py: Module is used to implement cable section analysis and an event study
"""

__author__ = "Chakraborty, S."
__copyright__ = ""
__credits__ = []
__license__ = "MIT"
__version__ = "1.0."
__maintainer__ = "Chakraborty, S."
__email__ = "chakras4@erau.edu"
__status__ = "Research"



#!/usr/bin/env python

"""
    datasets.py: This is a data holder module with various datatypes
    1. Layer: Contains individual layer's conductivity and thickness <1D>
    2. Site: Contains sites with individual multiple layer
    3. Holds multiple types of sites/layers
"""

__author__ = "Chakraborty, S."
__copyright__ = ""
__credits__ = []
__license__ = "MIT"
__version__ = "1.0."
__maintainer__ = "Chakraborty, S."
__email__ = "shibaji7@vt.edu"
__status__ = "Research"

from types import SimpleNamespace

import numpy as np
import pandas as pd
from scipy import constants as C


class Layer(object):
    """ """

    def __init__(self, name, tickness, conductivity):
        self.name = name
        self.thickness = tickness
        self.conductivity = conductivity
        self.resistivity = 1.0 / conductivity
        return


class Site(object):
    def __init__(self, layers, desciption, name):
        self.layers = layers
        self.desciption = desciption
        self.name = name
        return

    def get_thicknesses(self, index=-1):
        th = [l.thickness for l in self.layers]
        th = th[index] if index >= 0 else th
        return th

    def get_conductivities(self, index=-1):
        cd = [l.conductivity for l in self.layers]
        cd = cd[index] if index >= 0 else cd
        return cd

    def get_resistivities(self, index=-1):
        rv = [l.resistivity for l in self.layers]
        rv = rv[index] if index >= 0 else rv
        return rv

    def get_names(self, index=-1):
        na = [l.name for l in self.layers]
        na = na[index] if index >= 0 else na
        return na

    def get(self, index=-1):
        o = pd.DataFrame()
        o["names"] = self.get_names(index)
        o["conductivities"] = self.get_conductivities(index)
        o["resistivities"] = self.get_resistivities(index)
        o["thicknesses"] = self.get_thicknesses(index)
        return o

    def set_thicknesses(self, t, index=0):
        self.layers[index].thickness = t
        return self

    def calcZ(self, freqs, layer=0, return_all_layers=False):
        freqs = np.asarray(freqs)
        resistivities = np.asarray(self.get_resistivities())
        thicknesses = np.asarray(self.get_thicknesses())
        n = len(resistivities)
        nfreq = len(freqs)
        omega = 2 * np.pi * freqs
        complex_factor = 1j * omega * C.mu_0

        k = np.sqrt(1j * omega[np.newaxis, :] * C.mu_0 / resistivities[:, np.newaxis])
        Z = np.zeros(shape=(n, nfreq), dtype=complex)
        with np.errstate(divide="ignore", invalid="ignore"):
            Z[-1, :] = complex_factor / k[-1, :]
            r = np.zeros(shape=(n, nfreq), dtype=complex)
            for i in range(n - 2, -1, -1):
                r[i, :] = (1 - k[i, :] * Z[i + 1, :] / complex_factor) / (
                    1 + k[i, :] * Z[i + 1, :] / complex_factor
                )
                Z[i, :] = (
                    complex_factor
                    * (1 - r[i, :] * np.exp(-2 * k[i, :] * thicknesses[i]))
                    / (k[i, :] * (1 + r[i, :] * np.exp(-2 * k[i, :] * thicknesses[i])))
                )
        if freqs[0] == 0.0:
            Z[:, 0] = 0.0
        layer = layer if layer else 0
        Z_output = np.zeros(shape=(4, nfreq), dtype=complex)
        #################################################################
        # This factor '(1.0e-3 / C.mu_0)' converts K(f) and Z(f),
        # multiplying this factor makes Z -> K, devide this factor
        # regain Z.
        #################################################################
        Z_output[1, :] = Z[layer, :] * (1.0e-3 / C.mu_0)
        Z_output[2, :] = -Z_output[1, :]
        if return_all_layers:
            return Z_output, Z
        else:
            return Z_output

    def calcP(self, freqs, layer=0, return_Z=False):
        """
        Calculate the skin depth of the model
        """
        freqs = np.asarray(freqs)
        omega = 2 * np.pi * freqs

        if return_Z:
            Z_output, Z = self.calcZ(freqs=freqs, layer=layer, return_all_layers=True)
        else:
            Z_output = self.calcZ(freqs=freqs, layer=layer, return_all_layers=False)
        #################################################################
        # This factor '(1.0e-3 / C.mu_0)' converts K(f) and Z(f),
        # multiplying this factor makes Z -> K, devide this factor
        # regain Z.
        #################################################################
        p = Z_output[1, :] / (1j * omega * C.mu_0) / (1.0e-3 / C.mu_0)
        if return_Z:
            return p, Z_output, Z
        else:
            return p

    @staticmethod
    def init(conductivities, thicknesses, names, desciption, site_name):
        layers = []
        for c, t, n in zip(conductivities, thicknesses, names):
            layers.append(Layer(n, t, c))
        return Site(layers, desciption, site_name)


PROFILES = SimpleNamespace(
    **dict(
        BM=Site.init(
            conductivities=[0.2, 0.0003333, 0.02, 0.1, 1.12201],
            thicknesses=[2000, 75000, 332000, 250000, np.inf],
            names=[
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Lower Mantle",
            ],
            desciption="This model is Ben's model",
            site_name="Ben's Model",
        ),
        OM=Site.init(
            conductivities=[3.3333333, 0.2, 0.0003333, 0.02, 0.1, 1.12201],
            thicknesses=[1000, 2000, 75000, 332000, 250000, np.inf],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Lower Mantle",
            ],
            desciption="This model is modified Ben's model. An 1 km thick ocean on top of Ben's model.",
            site_name="Ocean Model",
        ),
        DB=Site.init(
            conductivities=[0.00005, 0.005, 0.001, 0.01, 0.3333333],
            thicknesses=[15000, 10000, 125000, 200000, np.inf],
            names=[
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Lower Mantle",
            ],
            desciption="This model is suggested by David.",
            site_name="David's Model",
        ),
        Quebec=Site.init(
            conductivities=[0.00005, 0.005, 0.001, 0.01, 0.3333333],
            thicknesses=[15000, 10000, 125000, 200000, np.inf],
            names=[
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Lower Mantle",
            ],
            desciption="This model is suggested by David.",
            site_name="David's Model",
        ),
        UN=Site.init(
            conductivities=[0.2, 0.2, 0.2, 0.2, 1.12201],
            thicknesses=[2000, 75000, 332000, 250000, np.inf],
            names=[
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Lower Mantle",
            ],
            desciption="This model is suggested by David.",
            site_name="Uniform Model",
        ),
        CS=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[100, 3000, 20000, 140000, 246900, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Used in IGS study by David",
            site_name="Continental Shelf",
        ),
        SO=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[1000, 2000, 10000, 70000, 327000, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Used in IGS study by David",
            site_name="Shallow Ocean",
        ),
        DO=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[4000, 2000, 10000, 70000, 327000, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Used in IGS study by David",
            site_name="Deep Ocean",
        ),
        LD=Site.init(
            conductivities=[0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[1000, 20000, 140000, 249000, 250000, 340000],
            names=[
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Used in IGS study by David",
            site_name="Land",
        ),
        CS_W=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[100, 8000, 15000, 150000, 236900, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Continental Shelf West",
        ),
        DO_1=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[4000, 4000, 10000, 145000, 247000, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Deep Ocean",
        ),
        DO_2=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[5200, 2000, 10000, 140000, 252800, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Deep Ocean",
        ),
        DO_3=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[4000, 2000, 10000, 140000, 254000, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Deep Ocean",
        ),
        DO_4=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[4800, 1000, 10000, 70000, 324200, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Deep Ocean",
        ),
        DO_5=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[4000, 1000, 10000, 70000, 324200, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Deep Ocean",
        ),
        MAR=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[3000, 0, 10000, 25000, 372000, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Mid-Atlantic Ridge",
        ),
        DO_6=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[4500, 1500, 10000, 70000, 324000, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Deep Ocean",
        ),
        CS_E=Site.init(
            conductivities=[3.3333333, 0.3333333, 0.00033333333, 0.001, 0.01, 0.1, 1],
            thicknesses=[100, 3000, 20000, 120000, 266900, 250000, 340000],
            names=[
                "Seawater",
                "Sediments",
                "Crust",
                "Lithosphere",
                "Upper Mantle",
                "Transition Zone",
                "Lower Mantle",
            ],
            desciption="Space Weather Paper: Cable (TAT-8) Section 1 / Used in SW 1989 Storm study",
            site_name="Continental Shelf East",
        ),
    )
)



"""utility.py: Module is used to implement utility methods.
"""

__author__ = "Chakraborty, S."
__copyright__ = ""
__credits__ = []
__license__ = "MIT"
__version__ = "1.0."
__maintainer__ = "Chakraborty, S."
__email__ = "shibaji7@vt.edu"
__status__ = "Research"

import math
from math import acos, asin, cos, radians, sin, sqrt
from types import SimpleNamespace

import numpy as np


class RecursiveNamespace(SimpleNamespace):
    """A simple object subclass that provides attribute access to its namespace.

    Methods:
        map_entry
    """

    @staticmethod
    def map_entry(entry: dict) -> dict:
        """Methods to convert a dictionary to namespace

        Arguments:
            entry: `dict` to convert into `scubas.utils.RecursiveNamespace`

        Returns:
            If `dict` return `scubas.utils.RecursiveNamespace` else itself
        """
        if isinstance(entry, dict):
            return RecursiveNamespace(**entry)
        return entry

    def __init__(self, **kwargs):
        """Initialize the parameters provided by kwargs."""
        super().__init__(**kwargs)
        for key, val in kwargs.items():
            if type(val) == dict:
                setattr(self, key, RecursiveNamespace(**val))
            elif type(val) == list:
                setattr(self, key, list(map(self.map_entry, val)))
        return


def frexp102str(x: float) -> str:
    """Convert to a float to `str` in the form of mantesa-exponent.

    Arguments:
        x: `float` number to be converted.

    Returns:
        `str` value of the `float` data in mantesa-exponent form.
    """
    exp = int(math.log10(x))
    if exp <= 0:
        exp -= 1
    m, exp = x / 10**exp, exp
    txt = r"%.2f$\times 10^{%d}$" % (m, exp)
    return txt


def fft(X: np.array, dT: float, remove_zero_frequency: bool = True) -> tuple:
    """This method runs FFT using NumPy package of a real signal (X).

    Arguments:
        X: Timeseries data
        dT: Timedelta in seconds
        remove_zero_frequency: If true replace the first 0 frequency (DC component)

    Returns:
        FFT values with frequency as a tuple.
    """
    n = len(X)
    Y = 2.0 / n * np.fft.rfft(X)
    f = np.fft.rfftfreq(len(X)) / dT
    if remove_zero_frequency:
        f[0] = f[1]
    return (Y, f)


def ifft(Y: np.array) -> np.array:
    """This method runs Inverse FFT using NumPy package of a complex signal (Y).

    Arguments:
        Y: Complex FFT dataset

    Returns:
        Timeseries values.
    """
    n = len(Y)
    X = np.fft.irfft(Y) * 2 * n
    return X


def component_mappings(field="B2E", comp="X"):
    """
    This method holds components mapping from
    (i) B2E
    """
    _map_ = {"B2E": {"X": "Y", "Y": "X"}}
    return _map_[field][comp]


def component_sign_mappings(fromto="BxEy"):
    """
    This method holds components mapping from
    (i) B2E sign change (Bx to Ey) and (By to Ex)
    """
    _map_ = {
        "BxEy": -1.0,
        "ByEx": 1.0,
    }
    return _map_[fromto]


class GreatCircle(object):
    """An object to calculate Great Circle distances (km) between two location given by latitude/longitude

    Methods:
        great_circle
        haversine
    """

    def __init__(self, initial, final, Re=6371.0):
        """Initialize the location points"""
        self.initial = initial
        self.final = final
        self.Re = Re
        return

    def check_location(self, loc):
        """
        Check lat/lon exists in file or not
        """
        tag = True if (hasattr(loc, "lat") and hasattr(loc, "lon")) else False
        return tag

    def great_circle(self, initial=None, final=None):
        """The Great Circle distance formula computes the shortest distance
        path of two points on the surface of the sphere. That means, when applies
        this to calculate distance of two locations on Earth, the formula assumes
        that the Earth is spherical.
        """
        initial = initial if initial else self.initial
        final = final if final else self.final
        if self.check_location(initial) and self.check_location(final):
            lon1, lat1, lon2, lat2 = map(
                radians,
                [self.initial.lon, self.initial.lat, self.final.lon, self.final.lat],
            )
            return self.Re * (
                acos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(lon1 - lon2))
            )
        else:
            return np.nan

    def haversine(self, initial=None, final=None):
        """Haversine formula is also another formula to calculate distance of two
        locations on a sphere using the law of haversine.
        """
        initial = initial if initial else self.initial
        final = final if final else self.final
        if self.check_location(initial) and self.check_location(final):
            lon1, lat1, lon2, lat2 = map(
                radians,
                [self.initial.lon, self.initial.lat, self.final.lon, self.final.lat2],
            )
            dlon = lon2 - lon1
            dlat = lat2 - lat1
            a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
            return 2 * self.Re * asin(sqrt(a))
        else:
            return np.nan



"""
    current_sources.py: Module is used to implement ionosphereic 
                        current sources and associated ground magnetic 
                        purturbations.
"""

__author__ = "Chakraborty, S."
__copyright__ = ""
__credits__ = []
__license__ = "MIT"
__version__ = "1.0."
__maintainer__ = "Chakraborty, S."
__email__ = "chakras4@erau.edu"
__status__ = "Research"

from enum import Enum

import numpy as np
import scipy.constants as C
from loguru import logger

from scubas.datasets import PROFILES, Site


class Direction(Enum):
    GEO_NORTH_SOUTH = 1
    GEO_EAST_WEST = 2
    GEO_VERT_DOWN = 3

    @classmethod
    def enumerate_key(cls, parameter: str = "B"):
        """
        Enumerate possible keys based on parameter
        """
        if cls == Direction.GEO_EAST_WEST:
            key = rf"${parameter}$_y"
        if cls == Direction.GEO_NORTH_SOUTH:
            key = rf"${parameter}$_x"
        if cls == Direction.GEO_VERT_DOWN:
            key = rf"${parameter}$_z"
        return key


class Current(object):
    def __init__(
        self,
        h: float = 110.0,
        a: float = 0.0,
        I: float = 1.0,
        p: np.array = 0.0 * np.arange(10, 200),
        omega: np.array = 2 * np.pi * np.arange(10, 200),
        site: Site = PROFILES.Quebec,
        length_unit: str = "km",
        current_unit: str = "MA",
        current_direction: Direction = Direction.GEO_EAST_WEST,
        t: int = 0,
        layer: int = 0,
        return_Z: bool = False,
    ):
        """
        Current systems for one time instance / one location
            h: Height given in unit
            a: Width of the current, 'Cauchy Distributed'
            I: Currents in Amps/M Amps
            p: List of skin depths of the wave Z/(j*omega)
            omega: List of angular frequencies in rad/s
            site: Site profile default Quebec
            length_unit: Units for h, p, and a
            current_unit: Units of I
            current_direction: Direction of the currents, in geographic coordinates
            t: Time instance of the current
            layer: Layer where calculate Z and p
        """
        self.length_unit = length_unit
        self.current_unit = current_unit
        self.length_unit_multiplier = 1e3 if length_unit == "km" else 1.0
        self.current_unit_multiplier = 1e6 if current_unit == "MA" else 1.0
        self.h = h * self.length_unit_multiplier  # in meters
        self.p = p * self.length_unit_multiplier  # in meters
        self.a = a * self.length_unit_multiplier  # in meters
        self.I = I * self.current_unit_multiplier  # in Amps
        self.omega = omega
        self.freq = omega / (2 * np.pi)
        self.site = site
        self.current_direction = current_direction
        self.t = t
        # If site is given calculate the skin depth p at top layer
        if site:
            ret = self.site.calcP(self.freq, layer=layer, return_Z=return_Z)
            if return_Z:
                self.p, self.Z_out, self.Z = ret
            else:
                self.p = ret
        return

    def compute_B_field(
        self,
        x: np.array = np.zeros((1)),
        length_unit: str = "km",
    ):
        """
        For a line current along y directed, we can only expect
        magnetic field along x and z directed.
        x: List of distance along x-direction given in unit
        length_unit: Units for x
        """
        length_unit_multiplier = 1e3 if length_unit == "km" else 1.0
        x *= length_unit_multiplier  # to meters
        B_field_directions = [
            Direction.GEO_EAST_WEST,
            Direction.GEO_NORTH_SOUTH,
            Direction.GEO_VERT_DOWN,
        ]
        # Remove 'y' directions only from the magnetic fields
        B_field_directions.remove(self.current_direction)
        logger.info(f"|Current| along : {self.current_direction} is {self.I} A")
        logger.info(
            f"Direction of the produced magnetic fields: {[d for d in B_field_directions]}"
        )
        # Modified height by Cauchy distributed current source
        h, h_mod, factor = (
            self.a + self.h,
            self.a + self.h + (2 * self.p),
            (C.mu_0 * self.I / (2 * np.pi)),
        )
        self.B_fields = [
            dict(
                direction=B_field_directions[0],
                total=factor
                * ((h / (h**2 + x**2)) + (h_mod / (h_mod**2 + x**2))),
                external=factor * (h / (h**2 + x**2)),
                internal=factor * (h_mod / (h_mod**2 + x**2)),
            ),
            dict(
                direction=B_field_directions[1],
                total=-1
                * factor
                * ((x / (h**2 + x**2)) - (x / (h_mod**2 + x**2))),
                external=-1 * factor * (x / (h**2 + x**2)),
                internal=factor * (x / (h_mod**2 + x**2)),
            ),
        ]
        for b in self.B_fields:
            logger.warning(
                f"Magnetic field B along {b['direction']} is {b['total'].min()},{b['total'].max()}"
            )
        self.__compute_E_field__(x, "m")
        self.B_field_units = "T"
        return

    def __compute_E_field__(
        self,
        x: np.array = np.zeros((1)),
        length_unit: str = "km",
    ):
        """
        For a line current along y directed, we can only expect
        magnetic field along y directed.
        """
        length_unit_multiplier = 1e3 if length_unit == "km" else 1.0
        x *= length_unit_multiplier  # to meters
        E_field_directions = [self.current_direction]
        logger.info(f"Direction of the E_field: {E_field_directions}")
        # Modified height by Cauchy distributed current source
        h = self.a + self.h
        self.E_fields = [
            dict(
                direction=E_field_directions[0],
                total=-1j
                * (self.omega * C.mu_0 * self.I / (2 * np.pi))
                * np.log(
                    np.sqrt((h + 2 * self.p) ** 2 + x**2) / np.sqrt(h**2 + x**2)
                ),
            ),
        ]
        for e in self.E_fields:
            logger.warning(
                f"Magnetic field E along {e['direction']} is {e['total'].min()},{e['total'].max()}"
            )
        self.E_field_units = "V/m"
        return



# read the contents of your README file
from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="scubas",
    version="0.1.6",
    packages=find_packages(),
    package_dir={"scubas": "scubas"},
    package_data={"scubas": []},
    include_package_data=True,
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    author="Shibaji Chakraborty",
    author_email="shibaji7@vt.edu",
    maintainer="Shibaji Chakraborty",
    maintainer_email="shibaji7@vt.edu",
    license="MIT License",
    description="SCUBAS: Submarine Cables Upset by Auroral Streams.",
    # long_description="Model compute electrical surges in submarine cables induced by geomagnetic activities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "pandas",
        "matplotlib>=3.3.4",
        "pyproj",
        "loguru",
        "scipy",
        "SciencePlots",
    ],
    keywords=[
        "Python>=3.6",
        "submarine cable",
        "electrical surge",
        "geomagnetic induction",
        "geomagnetic disturbance",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentations": "https://scubas.readthedocs.io/en/latest/",
    },
    url="https://github.com/shibaji7/SCUBAS",
)



"""
    model.py: Module is used to implement Ocean Model with Layered-Earth structures.
    Class:
    ------
    OceanModel: Dedicated for ocean - Earth B and E field transfer functions
"""

__author__ = "Chakraborty, S."
__copyright__ = ""
__credits__ = []
__license__ = "MIT"
__version__ = "1.0."
__maintainer__ = "Chakraborty, S."
__email__ = "shibaji7@vt.edu"
__status__ = "Research"


import datetime as dt

import numpy as np
import pandas as pd
from loguru import logger
from scipy import constants as C

from scubas.utils import component_mappings, component_sign_mappings, fft, ifft


class OceanModel(object):
    """
    This class is to emulate electric and magnetic fields
    in a 1D ocean. We provide the 1D Earth conductivity model and
    one layer of ocean parameters (depth and resistivity) to emulate
    1D ocean.
    """

    def __init__(
        self,
        site,
        flim=[1e-6, 1e0],
    ):
        """
        Initialize the ocean model
        """
        logger.info(f"Compile Ocean-model: {site.name} E- and B-Fields")
        self.site = site
        self.flim = flim
        self.freqs = np.linspace(flim[0], flim[1], int(flim[1] / flim[0]) + 1)
        self.functions = {
            "E2B": lambda Z, Zd, kd: Zd / (np.cosh(kd) + (Zd * np.sinh(kd) / Z)),
            "B2B": lambda Z, Zd, kd: 1.0 / (np.cosh(kd) + (Zd * np.sinh(kd) / Z)),
        }
        return

    def calcZ(self, freqs=None):
        """
        Compute the Km(f) for different layers
        """
        freqs = np.copy(self.freqs) if freqs is None else freqs
        omega = 2 * C.pi * freqs
        sigma_s = 1 / self.site.layers[0].resistivity
        k2 = 1.0j * omega * C.mu_0 * sigma_s
        k = np.sqrt(k2)
        Zo = (1.0j * omega * C.mu_0 / k) / (C.mu_0 / 1.0e-3)
        Kf = self.site.calcZ(freqs, layer=1)[1]
        self.Z = {"ocean": Zo, "floor": Kf}
        return

    def calcTF(self, kinds=["E2B", "B2B"], freqs=None):
        """
        Calculate the transfer functions.
        """
        freqs = np.copy(self.freqs) if freqs is None else freqs
        self.calcZ(freqs)
        omega = 2 * C.pi * freqs
        Zd = self.Z["floor"]
        Z = self.Z["ocean"]
        TFs = {}

        omega = 2 * C.pi * freqs
        sigma_s = 1 / self.site.layers[0].resistivity
        k2 = 1.0j * omega * C.mu_0 * sigma_s
        k = np.sqrt(k2)
        kd = k * self.site.layers[0].thickness

        for kind in kinds:
            TFs[kind] = self.functions[kind](Z, Zd, kd)
        return TFs

    def get_TFs(self, key="E2B", freqs=None):
        """
        Fetch the transfer function based on key
        and frequency.
        """
        TFs = self.calcTF(freqs=freqs)
        tf = pd.DataFrame()
        tf["freq"], tf[key] = np.copy(self.freqs) if freqs is None else freqs, TFs[key]
        return tf

    def read_iaga(self, file, return_xyzf=True, return_header=False):
        """
        Read IAGA profiles
        """

        # Read Headers
        header_records = {"header_length": 0}

        with open(file, "r") as openfile:
            for newline in openfile:
                if newline[0] == " ":
                    header_records["header_length"] += 1
                    label = newline[1:24].strip()
                    description = newline[24:-2].strip()
                    header_records[label.lower()] = description

        if len(header_records["reported"]) % 4 != 0:
            raise ValueError(
                "The header record does not contain 4 values: {0}".format(
                    header_records["reported"]
                )
            )
        record_length = len(header_records["reported"]) // 4
        column_names = [
            x for x in header_records["reported"][record_length - 1 :: record_length]
        ]
        seen_count = {}
        for i, col in enumerate(column_names):
            if col in seen_count:
                column_names[i] += str(seen_count[col])
                seen_count[col] += 1
            else:
                seen_count[col] = 1
        df = pd.read_csv(
            file,
            header=header_records["header_length"],
            delim_whitespace=True,
            parse_dates=[[0, 1]],
            infer_datetime_format=True,
            index_col=0,
            usecols=[0, 1, 3, 4, 5, 6],
            na_values=[99999.90, 99999.0, 88888.80, 88888.00],
            names=["Date", "Time"] + column_names,
        )
        df.index.name = "Time"
        if return_xyzf and "X" not in column_names and "Y" not in column_names:
            # Convert the data to XYZF format
            # Only convert HD
            if "H" not in column_names or "D" not in column_names:
                raise ValueError(
                    "Only have a converter for HDZF->XYZF\n"
                    + "Input file is: "
                    + header_records["reported"]
                )

            # IAGA-2002 D is reported in minutes of arc.
            df["X"] = df["H"] * np.cos(np.deg2rad(df["D"] / 60.0))
            df["Y"] = df["H"] * np.sin(np.deg2rad(df["D"] / 60.0))
            del df["H"], df["D"]
        if return_header:
            return df, header_records
        else:
            return df

    def read_Bfield_data(self, files, return_xyzf=True, csv_file_date_name="Date"):
        """
        Read B-Files
        """
        self.Bfield = pd.DataFrame()
        for file in files:
            logger.info(f"Read file: {file}")
            file_type = file.split(".")[-1]
            logger.info(f"Reading file types: {file_type}")
            if file_type == "txt":
                o = self.read_iaga(file, return_xyzf, return_header=False)
            elif file_type == "csv":
                o = pd.read_csv(file, parse_dates=[csv_file_date_name])
                o = o.rename(columns={csv_file_date_name: "Time"})
                o = o.set_index("Time")
                o.index.name = "Time"
            self.Bfield = pd.concat([self.Bfield, o])
        return self.Bfield

    def to_Efields(self, Bfield=None, components=["X", "Y"], p=None):
        """
        Compute Et using numerical FFT and IFFT block
        Bfield: Dataframe containing B-field
        components: [X and Y] for B-fields
        """
        self.Bfield = Bfield if Bfield else self.Bfield
        self.components = components
        self.Efield = pd.DataFrame()
        self.Efield["Time"] = self.Bfield.index.tolist()
        tx = self.Efield.Time.tolist()[0]
        if isinstance(tx, dt.datetime) or isinstance(tx, pd.Timestamp):
            t = np.array(self.Efield.Time.apply(lambda x: (x - tx).total_seconds()))
            self.Efield["dTime"], self.Bfield["dTime"] = t, t
        else:
            t = np.array(self.Efield.Time)
            self.Efield["dTime"], self.Bfield["dTime"] = (t[1] - t[0], t[1] - t[0])
        for a in self.components:
            Bt = np.array(self.Bfield[a])
            proc = Preprocess(np.array(self.Bfield.dTime), np.array(Bt))
            Bt = proc.detrend_magnetic_field(p)
            dT = (
                self.Bfield.dTime.tolist()[1]
                if "dTime" in self.Bfield.columns
                else t[1] - t[0]
            )
            Bf, f = fft(Bt, dT)
            E2B = np.array(self.get_TFs(freqs=f).E2B)
            Et = ifft(
                component_sign_mappings(
                    "B%sE%s" % (a.lower(), component_mappings("B2E", a).lower())
                )
                * E2B
                * Bf
            )
            self.Efield[component_mappings("B2E", a)] = Et
        self.Efield = self.Efield.set_index("Time")
        self.components = [component_mappings("B2E", a) for a in self.components]
        return


class Preprocess(object):
    """
    This class is used to process the B-field and E-field data.
    """

    def __init__(self, t, field, p=0.1):
        """
        Parameters:
        -----------
        t: Array of time seconds
        field: Array containing the data
        p: Tapering value (0-1)
        """
        self.t = t
        self.field = field
        self.p = p
        return

    def get_tapering_function(self, p=None):
        """
        This method is resposible for generateing
        tapering function based on time sequence t
        and tapering coefficient p
        """
        p = self.p if p is None else p
        T = len(self.t)
        P, P2 = int(T * p), int(T * p / 2)
        w = np.zeros_like(self.t)
        w[:P2] = 0.5 * (1 - np.cos(2 * np.pi * self.t[:P2] / P))
        w[P2 : T - P2] = 1.0
        w[T - P2 :] = 0.5 * (
            1 - np.cos(2 * np.pi * (self.t[-1] - self.t[T - P2 :]) / P)
        )
        return w

    def detrend_magnetic_field(self, p=None):
        """
        This method is resposible for detrend
        magnetic field data and taper it to reduce
        spurious frequency components.
        """
        p = self.p if p is None else p
        fmed = np.median(self.field[:120])
        w = self.get_tapering_function(p)
        f = (self.field - fmed) * w
        return f
